# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import asyncio
import io
import json
import os
import random
import sqlite3
import string
import time
from os import path
from tempfile import TemporaryFile
from typing import Any, BinaryIO, List, Optional, Tuple, Union

import cloudpickle as pickle
import requests
from aiohttp import ClientSession
from app.core.api import DataService
from app.core.config import settings
from app.core.db import Database
from app.core.get_svc_uri import DataURI
from app.schemas.common import HTTPExceptionSchema
from app.schemas.workflow import (
    DeleteResultResponse,
    InsertResultResponse,
    Node,
    Result,
    ResultFormats,
    UpdateResultResponse,
)
from fastapi import APIRouter, File, HTTPException, Query, Request, Response, UploadFile
from fastapi.responses import FileResponse, StreamingResponse

from covalent._shared_files.utils import encode_result

router = APIRouter()

data_svc = DataService()
db = Database()


# @router.middleware("http")
# TODO: figure out why the middleware doesn't work
# async def log_requests(request: Request, call_next):
#    idem = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
#    logger.info(f"rid={idem} start request path={request.url.path}")
#    start_time = time.time()
#
#    response = await call_next(request)
#
#    process_time = (time.time() - start_time) * 1000
#    formatted_process_time = "{0:.2f}".format(process_time)
#    logger.info(
#        f"rid={idem} completed_in={formatted_process_time}ms status_code={response.status_code}"
#    )
#
#    return response


async def _get_result_file(dispatch_id: str) -> bytes:
    filename = _get_result_from_db(dispatch_id, "results_filename")
    path = _get_result_from_db(dispatch_id, "results_path")
    if not dispatch_id or not filename or not path:
        raise HTTPException(status_code=404, detail="Result was not found")
    file = await data_svc.download(filename)
    return file


async def _upload_file(result_pkl_file: BinaryIO):
    results_object = {}
    dispatch_id = ""
    length = result_pkl_file.seek(0, 2)
    result_pkl_file.seek(0)
    try:
        results_object = pickle.load(result_pkl_file)
        dispatch_id = results_object.dispatch_id
        assert length > 0
    except:
        raise HTTPException(status_code=422, detail="Error in upload body.")
    result_pkl_file.seek(0)
    response = await data_svc.upload(result_pkl_file)
    filename = response.get("filename")
    path = response.get("path")
    error_detail = "Error in response from data service. " + str(response)
    if filename and path:
        if _add_record_to_db(dispatch_id, filename, path):
            return {"dispatch_id": dispatch_id}
        else:
            error_detail = "Error adding record to database."
    raise HTTPException(status_code=500, detail="Error adding record to database.")


async def _concurrent_download_and_serialize(semaphore, file_name, session):
    async with semaphore:
        # logger.debug(f"Downloading & serializing: {file_name}...")
        async with session.get(
            DataURI().get_route("/fs/download"), params={"file_location": file_name}
        ) as resp:
            result_binary = await resp.read()
            result = pickle.loads(result_binary)
            result_json = json.loads(encode_result(result))
            return result_json


def _get_results_from_db() -> List[Tuple[str, str]]:
    rows = db.value("SELECT id, results_filename FROM workflow")

    return rows


def _get_result_from_db(dispatch_id: str, field: str) -> Optional[str]:
    sql = f"SELECT {field} FROM workflow WHERE id = '{dispatch_id}'"
    value = db.value(sql)
    return value


def _add_record_to_db(dispatch_id: str, filename: str, path: str) -> None:
    if _get_result_from_db(dispatch_id, "results_filename"):
        sql = (
            "UPDATE workflow "
            f"SET results_filename = '{filename}', results_path = '{path}' "
            f"WHERE id = '{dispatch_id}' "
        )
    else:
        sql = f"INSERT INTO workflow (id, results_filename, results_path) VALUES('{dispatch_id}','{filename}','{path}')"

    response = db.value(sql)
    return response


@router.get(
    "/results",
    status_code=200,
    responses={
        200: {
            "model": List[Result],
            "description": "Return JSON serialized result objects from the database",
        },
    },
)
async def get_results(format: ResultFormats = ResultFormats.JSON) -> Any:
    """
    Return JSON serialized result objects from the database
    """
    results = _get_results_from_db()
    tasks = []
    # Set 10 concurrent requests at a time
    semaphore = asyncio.Semaphore(10)
    async with ClientSession() as session:
        # Send requests in parallel to Data Service for download
        for result in results:
            dispatch_id, file_name = result
            task = asyncio.create_task(
                _concurrent_download_and_serialize(semaphore, file_name, session)
            )
            tasks.append(task)
        # Wait until parallel tasks return and gather results
        responses = await asyncio.gather(*tasks, return_exceptions=False)
        return responses


@router.get(
    "/results/{dispatch_id}",
    status_code=200,
    response_class=FileResponse,
    responses={
        404: {"model": HTTPExceptionSchema, "description": "Result was not found"},
        200: {
            "content": {"application/octet-stream": {}},
            "description": "Return binary content of file.",
        },
    },
)
async def get_result(*, dispatch_id: str, format: ResultFormats = ResultFormats.BINARY) -> Any:
    """
    Get a result object as pickle file
    """
    result_binary: bytes = await _get_result_file(dispatch_id)
    if format == ResultFormats.JSON:
        result = pickle.loads(result_binary)
        result_json_stringified = encode_result(result)
        return Response(content=result_json_stringified, media_type="application/json")
    else:
        return StreamingResponse(io.BytesIO(result_binary), media_type="application/octet-stream")


@router.post("/results", status_code=200, response_model=InsertResultResponse)
async def insert_result(
    *,
    result_pkl_file: UploadFile,
) -> Any:
    """
    Submit pickled result file
    """

    response = await _upload_file(result_pkl_file.file)
    return response


@router.put(
    "/results/{dispatch_id}",
    status_code=200,
    responses={
        404: {"model": HTTPExceptionSchema, "description": "Result was not found"},
        200: {
            "model": UpdateResultResponse,
            "description": "Return message indicating success of updating task",
        },
    },
)
async def update_result(*, dispatch_id: str, task: bytes = File(...)) -> Any:
    """
    Update a result object's task
    """
    result = await _get_result_file(dispatch_id)
    results_object = pickle.loads(result)
    task = pickle.loads(task)
    results_object._update_node(**task)

    pickled_result = io.BytesIO(pickle.dumps(results_object))
    uploaded = await _upload_file(pickled_result)

    if uploaded:
        return {"response": "Task updated successfully"}


@router.delete("/results", status_code=200, response_model=DeleteResultResponse)
def delete_result(*, dispatch_ids: List[str] = Query([])) -> DeleteResultResponse:
    # Retrieve file paths from db
    filenames = []

    # TODO: add batch method to request set of results
    for dispatch_id in dispatch_ids:
        filenames += [_get_result_from_db(dispatch_id, "results_filename")]

    # Request deletion from the data service
    r = requests.delete(DataURI().get_route("/fs/delete"), params={"obj_names": filenames})

    deleted_results = r.json()["deleted"]

    deleted_ids = []
    failed_ids = []
    if len(dispatch_ids) > 1:
        sql = f"DELETE FROM workflow WHERE id IN {tuple(dispatch_ids)}"
    else:
        sql = f"DELETE FROM workflow WHERE id = '{dispatch_ids[0]}'"
    rows_affected = db.value(sql)

    return {
        "deleted": rows_affected,
        "failed": len(dispatch_ids) - rows_affected,
    }



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
from pathlib import Path

# Creating a custom logger
logger = logging.getLogger(__name__)

# File handler for saving logs
Path("logs").mkdir(exist_ok=True)
file_handler = logging.FileHandler("logs/runner_logs.log")
file_handler.setLevel(logging.WARNING)

# Creating the file formatter and adding it to the file handler
file_formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
file_handler.setFormatter(file_formatter)

# Add handlers to the logger
logger.addHandler(file_handler)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


import sys
from datetime import datetime, timezone
from multiprocessing import Queue as MPQ

import cloudpickle as pickle
from app.core.execution import (
    cancel_running_task,
    generate_task_result,
    get_task_status,
    run_tasks_with_resources,
    send_task_update_to_dispatcher,
)
from app.core.runner_logger import logger
from app.schemas.task import CancelResponse, RunTaskResponse, TaskPickleList, TaskStatus
from fastapi import APIRouter, File

from covalent._results_manager.result import Result

INITAL_AVAILABLE_RESOURCES = 1

print(f"INITIAL_AVAILABLE_RESOURCES: {INITAL_AVAILABLE_RESOURCES}")

resources = MPQ()
resources.put(INITAL_AVAILABLE_RESOURCES)


router = APIRouter()

# Sample:
# example_ultimate_data = {

#     "dispatch_id_1": {
#         "task_id_1": TaskData(None, None, None),
#         "task_id_2": TaskData(None, None, None),
#     },

#     "dispatch_id_2": {
#         "task_id_1": TaskData(None, None, None),
#         "task_id_2": TaskData(None, None, None),
#     }

# }

ultimate_dict = {}

logger.warning("Runner Service Started")


@router.post("/{dispatch_id}/tasks", status_code=202, response_model=RunTaskResponse)
async def run_tasks(*, dispatch_id: str, tasks: bytes = File(...)) -> RunTaskResponse:
    """
    Run a list of tasks

    Note: The request body contains a list of "string"s which are pickled objects (bytes) containing
    the following info about the Task:

    `task_id`: Task ID \n
    `func`: Callable function which will be run \n
    `args`: Positional arguments for `func` \n
    `kwargs`: Keyword arguments for `func` \n
    `executor`: Executor object (https://covalent.readthedocs.io/en/latest/api/api.html#local-executor)\n

    Example: \n

    ```
    def task_func_1(x, y):
        return x * y

    def task_func_2(a):
        return a ** 2

    executor_1 = LocalExecutor() \n
    executor_2 = LocalExecutor()

    task_1 = pickle.dumps({"task_id": 0, "func": task_func_1, "args": (1, 2), "kwargs": {}, "executor": executor_1}) \n
    task_2 = pickle.dumps({"task_id": 1, "func": task_func_2, "args": (3,), "kwargs": {}, "executor": executor_2})

    requests.post(f'localhost:48008/api/workflow/{dispatch_id}/tasks', body=pickle.dumps([task_1, task_2]))
    ```
    """

    logger.warning(f"POST on /{dispatch_id}/tasks called to submit a list of tasks")

    runnable_tasks = pickle.loads(tasks)

    logger.warning(f"ultimate dict before {ultimate_dict}")

    tasks_left_to_run = run_tasks_with_resources(
        dispatch_id, runnable_tasks, resources, ultimate_dict
    )

    logger.warning(f"ultimate dict after {ultimate_dict}")

    # Returning the task ids which were not run due to insufficient resources
    left_out_task_ids = [task["task_id"] for task in tasks_left_to_run]

    return {"left_out_task_ids": left_out_task_ids}


@router.get("/{dispatch_id}/task/{task_id}", status_code=200, response_model=TaskStatus)
async def check_status(*, dispatch_id: str, task_id: int) -> TaskStatus:
    """
    Check status of a task
    """

    logger.warning(f"GET on /{dispatch_id}/task/{task_id} called to check status")

    # Pass in the executor and info_queue to get status from the executor
    task_status = get_task_status(
        pickle.loads(ultimate_dict[dispatch_id][task_id]["executor"]),
        ultimate_dict[dispatch_id][task_id]["info_queue"],
    )

    return {"status": f"{task_status}"}


@router.delete("/{dispatch_id}/task/{task_id}", status_code=200, response_model=CancelResponse)
async def cancel_task(*, dispatch_id: str, task_id: int) -> CancelResponse:
    """
    Cancel a task
    """

    print(f"DELETE on /{dispatch_id}/task/{task_id} called to cancel task", file=sys.stderr)

    # print(
    #     f"ultimate dict in cancel before cancelling {[(k, list(v)[0]) for k, v in ultimate_dict.items()]}",
    #     file=sys.stderr,
    # )

    # Cancel a task by calling its executor's cancel method and closing the info_queue
    cancel_running_task(
        executor=pickle.loads(ultimate_dict[dispatch_id][task_id]["executor"]),
        info_queue=ultimate_dict[dispatch_id][task_id]["info_queue"],
    )

    # Free the resources
    free_resources(dispatch_id=dispatch_id, task_id=task_id)

    # Terminate the process
    task_done(dispatch_id=dispatch_id, task_id=task_id)

    # Send updated task result to dispatcher
    task_result = generate_task_result(
        task_id=task_id,
        end_time=datetime.now(timezone.utc),
        status=Result.CANCELLED,
    )

    # ultimate_dict = {
    #     "dispatch_id": {"task_id": {"sdasf":12, "sfsdf": fdsf}}
    # }

    # print(
    #     f"ultimate dict in cancel after cancelling {[(k, list(v)[0]) for k, v in ultimate_dict.items()]}",
    #     file=sys.stderr,
    # )

    send_task_update_to_dispatcher(dispatch_id=dispatch_id, task_result=task_result)

    return {"cancelled_dispatch_id": f"{dispatch_id}", "cancelled_task_id": f"{task_id}"}


@router.post("/{dispatch_id}/task/{task_id}/free", status_code=200)
def free_resources(*, dispatch_id: str, task_id: int) -> None:
    """
    Callback to the runner to free resources
    """

    logger.warning(f"POST on /{dispatch_id}/task/{task_id}/free called to free resources")

    resources.put(resources.get() + 1)

    logger.warning("FREEING RESOURCES DONE")


@router.post("/{dispatch_id}/task/{task_id}/done", status_code=200)
def task_done(*, dispatch_id: str, task_id: int) -> None:
    """
    Callback to the runner to join the process
    """

    logger.warning(
        f"POST on /{dispatch_id}/task/{task_id}/done called to terminate and join the process"
    )

    ultimate_dict[dispatch_id][task_id]["process"].terminate()
    ultimate_dict[dispatch_id][task_id]["process"].join()

    del ultimate_dict[dispatch_id][task_id]

    if not ultimate_dict.get(dispatch_id):
        del ultimate_dict[dispatch_id]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


"""Workflow dispatch functionality."""

import sys
from datetime import datetime, timezone
from multiprocessing import Queue as MPQ
from queue import Empty
from typing import Dict, List, Tuple, Union

from app.core.dispatcher_logger import logger
from app.core.utils import (
    get_parent_id_and_task_id,
    is_empty,
    send_result_object_to_result_service,
    send_task_list_to_runner,
)

from covalent._results_manager import Result
from covalent._workflow.transport import _TransportGraph
from covalent.executor import BaseExecutor

from .utils import get_task_inputs, get_task_order, is_sublattice, preprocess_transport_graph


def dispatch_workflow(result_obj: Result, tasks_queue: MPQ) -> Result:
    """Responsible for starting off the workflow dispatch."""

    # logger.warning(f"Inside dispatch_workflow with dispatch_id {result_obj.dispatch_id}")

    if result_obj.status == Result.NEW_OBJ:
        result_obj = start_dispatch(result_obj=result_obj, tasks_queue=tasks_queue)

    elif result_obj.status == Result.COMPLETED:
        # TODO - Redispatch workflow for reproducibility

        pass

    elif result_obj.status == Result.CANCELLED:
        # TODO - Redispatch cancelled workflow

        # Change status to running if we want to redeploy this workflow
        result_obj._status = Result.RUNNING

        # TODO - Redispatch tasks

    elif result_obj.status == Result.FAILED:
        # TODO - Redispatch failed workflow

        pass

    return result_obj


def start_dispatch(result_obj: Result, tasks_queue: MPQ) -> Result:
    """Responsible for preprocessing the tasks, and sending the tasks for execution to the
    Runner API in batches. One of the underlying principles is that the Runner API doesn't
    interact with the Data API."""

    # logger.warning(f"Inside start_dispatch with dispatch_id {result_obj.dispatch_id}")

    # Change result status to running and set workflow execution start time.
    result_obj._status = Result.RUNNING
    result_obj._start_time = datetime.now(timezone.utc)

    # Initialize the result object
    result_obj = init_result_pre_dispatch(result_obj)

    # Send the initialized result to the result service
    send_result_object_to_result_service(result_obj)

    # Get the order of tasks to be run
    task_order: List[List] = get_task_order(result_obj)

    logger.warning(f"task_order: {task_order}")
    dispatch_runnable_tasks(result_obj, tasks_queue, task_order)

    send_result_object_to_result_service(result_obj)

    return result_obj


def dispatch_runnable_tasks(result_obj: Result, tasks_queue: MPQ, task_order: List[List]) -> None:
    """Get runnable tasks and dispatch them to the Runner API. Put the tasks that weren't picked
    up by the Runner API back in the queue."""

    print(
        f"Dispatch id: {result_obj.dispatch_id}, tasks order: {task_order}",
        file=sys.stderr,
    )

    val = tasks_queue.get()

    print(
        f"Dispatch id: {result_obj.dispatch_id}, tasks queue initial: {val}",
        file=sys.stderr,
    )

    tasks_queue.put(val)

    # To get the runnable tasks from first task order list
    # Sending the tasks_queue as well to handle the case of sublattices
    tasks, functions, input_args, input_kwargs, executors, next_tasks_order = get_runnable_tasks(
        result_obj=result_obj,
        tasks_order=task_order,
        tasks_queue=tasks_queue,
    )

    logger.warning(f"In dispatch_runnable_tasks with tasks: {tasks}")

    logger.warning(f"Set of next tasks to be run {next_tasks_order}")

    print(
        f"Dispatch id: {result_obj.dispatch_id}, will try to run these nodes: {tasks}",
        file=sys.stderr,
    )

    # The next set of tasks that can be run afterwards
    # This is the case of a new dispatch id in the list of dictionaries
    if next_tasks_order:
        if is_empty(tasks_queue):
            v = tasks_queue.get()
            logger.warning(f"LETS SEE IF TASKS QUEUE IS EMPTY ITS VAL IS: {v}")
            tasks_queue.put([{result_obj.dispatch_id: next_tasks_order}])
        else:

            all_tasks_lod = tasks_queue.get()

            # Checking if current dispatch id is a parent of last
            # dispatch id in the tasks queue
            last_dispatch_id = str(list(all_tasks_lod[0])[0])
            parent_id, _ = get_parent_id_and_task_id(last_dispatch_id)

            if parent_id == result_obj.dispatch_id:
                all_tasks_lod = all_tasks_lod + [{result_obj.dispatch_id: next_tasks_order}]
            else:
                all_tasks_lod = [{result_obj.dispatch_id: next_tasks_order}] + all_tasks_lod

            tasks_queue.put(all_tasks_lod)

    # Tasks which were not able to run
    unrun_tasks = run_tasks(
        results_dir=result_obj.results_dir,
        dispatch_id=result_obj.dispatch_id,
        task_id_batch=tasks,
        functions=functions,
        input_args=input_args,
        input_kwargs=input_kwargs,
        executors=executors,
    )

    print(f"Dispatch id: {result_obj.dispatch_id}, unrun_tasks: {unrun_tasks}", file=sys.stderr)

    # Add unrun tasks back to the tasks_queue
    final_task_order = tasks_queue.get()

    print(
        f"Dispatch id: {result_obj.dispatch_id}, intermediate_task_order: {final_task_order}",
        file=sys.stderr,
    )

    if unrun_tasks:
        if final_task_order is not None:
            print(f"unrun tasks: {unrun_tasks}")
            print(f"final task order: {final_task_order}")

            if final_task_order[0].get(result_obj.dispatch_id):
                final_task_order[0][result_obj.dispatch_id] = [unrun_tasks] + final_task_order[0][
                    result_obj.dispatch_id
                ]
            else:
                new_dict = {result_obj.dispatch_id: [unrun_tasks]}
                final_task_order = [new_dict] + final_task_order
        else:
            final_task_order = [{result_obj.dispatch_id: [unrun_tasks]}]

    logger.warning(f"Tasks which are yet to execute {final_task_order}")

    try:
        logger.warning(f"Before we put in final_task_order {tasks_queue.get_nowait()}")

    except Empty:
        logger.warning("Before we put in final_task_order, its empty")

    # Put the task order back into the queue
    tasks_queue.put(final_task_order)

    print(
        f"Dispatch id: {result_obj.dispatch_id}, tasks queue final: {final_task_order}",
        file=sys.stderr,
    )


def get_runnable_tasks(
    result_obj: Result,
    tasks_order: List[List],
    tasks_queue: MPQ,
) -> Tuple[List[int], List[bytes], List[List], List[Dict], List[BaseExecutor], List[List[int]]]:
    """Return a list of tasks that can be run and the corresponding executors and input
    parameters."""

    # logger.warning(f"In get_runnable_tasks task_order after get: {tasks_order}")

    # print(
    #     f"GET_RUNNABLE_TASKS - dispatch_id: {result_obj.dispatch_id}, tasks_order: {tasks_order}",
    #     file=sys.stderr,
    # )

    task_ids = tasks_order.pop(0)

    # print(
    #     f"GET_RUNNABLE_TASKS -  dispatch_id: {result_obj.dispatch_id}, task_ids: {task_ids}",
    #     file=sys.stderr,
    # )

    input_args = []
    input_kwargs = []
    executors = []
    runnable_tasks = []
    non_runnable_tasks = []
    functions = []

    for task_id in task_ids:

        task_name = result_obj._get_node_name(task_id)

        # Check whether task is of non-executable type
        result_obj, is_executable = preprocess_transport_graph(task_id, task_name, result_obj)

        if not is_executable:
            if task_id == task_ids[-1] and not runnable_tasks and tasks_order:
                return get_runnable_tasks(
                    result_obj=result_obj, tasks_order=tasks_order, tasks_queue=tasks_queue
                )
            continue

        serialized_function = result_obj.transport_graph.get_node_value(task_id, "function")

        # Get the task inputs from parents and edge names of this node
        task_inputs = get_task_inputs(task_id=task_id, node_name=task_name, result_obj=result_obj)

        executor = result_obj.transport_graph.get_node_value(task_id, "metadata")["executor"]

        # Check whether the node is a sublattice
        if is_sublattice(task_name):

            print("INSIDE SUBLATTICE", file=sys.stderr)

            # Get the sublattice
            sublattice = serialized_function.get_deserialized()

            # Build the graph for this sublattice
            sublattice.build_graph(*task_inputs["args"], **task_inputs["kwargs"])

            # Construct the result object for this sublattice
            sublattice_result_obj = Result(
                lattice=sublattice,
                results_dir=result_obj.lattice.metadata["results_dir"],
                dispatch_id=f"{result_obj.dispatch_id}:{task_id}",
            )

            print(
                f"TASKS ORDER IN SUBLATTICE CONDITION: {get_task_order(sublattice_result_obj)}",
                file=sys.stderr,
            )

            # Dispatch the sublattice recursively
            dispatch_workflow(result_obj=sublattice_result_obj, tasks_queue=tasks_queue)

        elif is_runnable_task(task_id, result_obj):

            # Add the details of this task to respective lists
            runnable_tasks.append(task_id)
            functions.append(serialized_function)
            input_args.append(task_inputs["args"])
            input_kwargs.append(task_inputs["kwargs"])
            executors.append(executor)

        else:
            # If the task is not runnable, i.e, its parents have not completed execution
            non_runnable_tasks.append(task_id)

    # If there are non runnable tasks then add them as well to next task order list else
    # only keep the node lists already present in tasks_order
    next_tasks_order = [non_runnable_tasks] + tasks_order if non_runnable_tasks else tasks_order

    # logger.warning(f"In get_runnable_tasks task_order after put: {next_tasks_order}")

    return runnable_tasks, functions, input_args, input_kwargs, executors, next_tasks_order


def init_result_pre_dispatch(result_obj: Result):
    """Initialize the result object transport graph before it is dispatched for execution."""
    result_obj._initialize_nodes()
    return result_obj


def run_tasks(
    results_dir: str,
    dispatch_id: str,
    task_id_batch: List[int],
    functions: List[bytes],
    input_args: List[List],
    input_kwargs: List[Dict],
    executors: List[Union[bytes, str]],
) -> List[int]:
    """Request Runner to execute tasks.

    The Runner might not have resources available to pick up the full batch of tasks. In that case,
    this function returns the list of task ids that were not picked up.
    """

    tasks_list = [
        {
            "task_id": task_id,
            "func": func,
            "args": args,
            "kwargs": kwargs,
            "executor": executor,
            "results_dir": results_dir,
        }
        for task_id, func, args, kwargs, executor in zip(
            task_id_batch, functions, input_args, input_kwargs, executors
        )
    ]

    return send_task_list_to_runner(dispatch_id=dispatch_id, tasks_list=tasks_list)


def is_runnable_task(task_id: int, result_obj: Result) -> bool:
    """Return status whether the task can be run based on whether the parent tasks have finished
    executing."""

    parent_node_ids: List[int] = result_obj.transport_graph.get_dependencies(task_id)

    return all(
        result_obj._get_node_status(node_id) == Result.COMPLETED for node_id in parent_node_ids
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Notification objects which can be attached to lattices and electrons."""

from abc import ABC, abstractmethod


class NotifyEndpoint(ABC):
    """Base notification endpoint class used for defining other types of
    notification endpoints. Subclasses must define a notify function which
    accepts a notification string forwarded to the endpoint. Other settings
    specific to the subclass should be passed in its constructor.
    """

    @abstractmethod
    def notify(self, message: str) -> None:  # pragma: no cover
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Amazon Simple Queue Service (SQS) implementation of the workflow queue."""

import json
import logging
import os
from enum import Enum
from typing import Callable

import botocore.exceptions
from aiobotocore.session import ClientCreatorContext, get_session
from app.core.config import settings

MQ_QUEUE_MESSAGE_GROUP_ID = settings.MQ_QUEUE_MESSAGE_GROUP_ID
MQ_QUEUE_MSG_WAIT_TIME = settings.MQ_QUEUE_MSG_WAIT_TIME

logger = logging.getLogger(__name__)


class AwsErrorCodes(Enum):
    NON_EXISTENT_QUEUE = "AWS.SimpleQueueService.NonExistentQueue"


class Queue:
    def __init__(
        self,
        queue_name: str = None,
        queue_url: str = None,
    ):
        self.queue_name = queue_name
        self.queue_url = queue_url

    def client_factory(self) -> ClientCreatorContext:
        """SQS client factory."""

        session = get_session()
        return session.create_client("sqs")

    async def get_queue_url(self) -> str:
        """Get the URL of the queue."""

        if self.queue_url:
            return self.queue_url

        async with self.client_factory() as client:
            try:
                response = await client.get_queue_url(QueueName=self.queue_name)
            except botocore.exceptions.ClientError as err:
                if err.response["Error"]["Code"] == AwsErrorCodes.NON_EXISTENT_QUEUE:
                    logging.error(f"Queue {self.queue_name} does not exist.")
                raise

            self.queue_url = response["QueueUrl"]
            return self.queue_url

    async def publish(
        self, message_body_dict: dict, message_group_id: str = MQ_QUEUE_MESSAGE_GROUP_ID
    ) -> dict:
        """Publish message to the queue.

        Args:
            message_body_dict: Message body takes the form {'dispatch_id': 'alpha_numeric_dispatch_id'}.
            message_group_id: Defaults to MQ_QUEUE_MESSAGE_GROUP_ID.

        Returns:
            Response from client when message is published.
        """

        async with self.client_factory() as client:
            queue_url = await self.get_queue_url()
            message_body_str = json.dumps(message_body_dict)
            response = await client.send_message(
                QueueUrl=queue_url,
                MessageBody=message_body_str,
                MessageGroupId=message_group_id,
            )
            return response

    async def poll_queue(self, message_handler: Callable) -> None:
        """Poll for messages published to the queue and apply the message handler.

        Args:
            message_handler:
                Function to apply to the message. The function takes a single input of a message body in json format.
        """

        async with self.client_factory() as client:
            queue_url = await self.get_queue_url()

            while True:
                response = await client.receive_message(
                    QueueUrl=queue_url, WaitTimeSeconds=MQ_QUEUE_MSG_WAIT_TIME
                )
                if "Messages" in response:
                    for msg in response["Messages"]:
                        msg_body = json.loads(msg["Body"])
                        try:
                            await message_handler(msg_body)
                            await client.delete_message(
                                QueueUrl=queue_url, ReceiptHandle=msg["ReceiptHandle"]
                            )
                        except Exception as err:
                            logging.error(err)
                            logging.exception(
                                f"Message {msg['ReceiptHandle']} was not able to be processed."
                            )



import os
from urllib.parse import urljoin

import requests
from app.core.config import settings


class APIService:
    def __init__(self, BASE_URI: str):
        if BASE_URI[len(BASE_URI) - 1] != "/":
            BASE_URI = f"{BASE_URI}/"
        self.BASE_URI = BASE_URI

    def _get_route(self, path: str):
        return urljoin(self.BASE_URI, path)

    def _format(self, response):
        return response.json()

    def post(self, path, json={}, params={}, data={}, files={}):
        route = self._get_route(path)
        return self._format(requests.post(route, json=json, params=params, data=data, files=files))

    def get(self, path, params={}):
        route = self._get_route(path)
        return self._format(requests.get(route, params=params))

    def delete(self, path, params={}):
        route = self._get_route(path)
        return self._format(requests.delete(route, params=params))

    def patch(self, path, json={}, params={}, data={}):
        route = self._get_route(path)
        return self._format(requests.patch(route, json=json, params=params, data=data))

    def put(self, path, json={}, params={}, data={}):
        route = self._get_route(path)
        return self._format(requests.put(route, json=json, params=params, data=data))


class ResultsService(APIService):
    def __init__(self):
        super().__init__(f"http://{settings.RESULTS_SVC_HOST}:{settings.RESULTS_SVC_PORT}")

    async def create_result(self, result_pkl_file: bytes):
        return self.post("api/v0/workflow/results", files={"result_pkl_file": result_pkl_file})



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import asyncio
import logging
import os

import requests
from app.core.config import settings
from app.core.dispatcher_logger import logger
from app.core.get_svc_uri import DispatcherURI
from app.core.queue import Queue
from dotenv import load_dotenv

logger = logging.getLogger(__name__)
load_dotenv()

MQ_QUEUE_NAME = settings.MQ_QUEUE_NAME


def send_dispatch_id(dispatch_id: str) -> None:
    """Submit workflow by sending dispatch id to Dispatcher service.

    Args:
        dispatch_id: identifier of the workflow to be submitted.
    """

    resp = requests.post(DispatcherURI().get_route(f"workflow/{dispatch_id}"))
    resp.raise_for_status()

    logger.warning(f"Dispatch id {dispatch_id} sent successfully.")


async def main():
    """Pick up workflows from the message queue and dispatch them one by one."""

    queue = Queue(queue_name=MQ_QUEUE_NAME)

    async def msg_handler(msg: dict) -> None:
        dispatch_id = msg["dispatch_id"]
        logger.info(f"Got dispatch_id: {dispatch_id} with type {type(dispatch_id)}")
        send_dispatch_id(dispatch_id=dispatch_id)

    try:
        await queue.poll_queue(message_handler=msg_handler)
    except Exception as err:
        logging.error(f"Queue consumer could not poll queue for messages: {err}.")
        raise


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        asyncio.ensure_future(main())
        loop.run_forever()
    except KeyboardInterrupt:
        loop.close()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

from pathlib import Path

from app.api.api_v0.api import api_router
from app.api.health import health_router
from app.core.config import settings
from fastapi import FastAPI

BASE_PATH = Path(__file__).resolve().parent

app = FastAPI(title="Covalent Runner Service API")


app.include_router(api_router, prefix=settings.API_V0_STR)
app.include_router(health_router)

if __name__ == "__main__":
    # Use this for debugging purposes only
    import uvicorn

    uvicorn.run(
        "main:app",
        host=settings.RUNNER_SVC_HOST,
        port=settings.RUNNER_SVC_PORT,
        log_level="debug",
        reload=True,
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Workflow result update functionality."""

import sys
from datetime import datetime, timezone
from multiprocessing import Queue as MPQ
from typing import Dict

from covalent._results_manager import Result

from .dispatch_workflow import dispatch_runnable_tasks
from .dispatcher_logger import logger
from .utils import (
    _post_process,
    are_tasks_running,
    generate_task_result,
    get_parent_id_and_task_id,
    get_result_object_from_result_service,
    is_empty,
    is_sublattice_dispatch_id,
    send_task_update_to_dispatcher,
    send_task_update_to_result_service,
)


def update_workflow_results(
    task_execution_results: Dict, dispatch_id: str, tasks_queue: MPQ
) -> Result:
    """Main update function. Called by the Runner API when there is an update for task
    execution status."""

    latest_result_obj: Result = get_result_object_from_result_service(dispatch_id=dispatch_id)

    # Update the task results
    latest_result_obj._update_node(**task_execution_results)
    send_task_update_to_result_service(dispatch_id, task_execution_results)

    if task_execution_results["status"] == Result.RUNNING:

        if is_sublattice_dispatch_id(latest_result_obj.dispatch_id):

            parent_dispatch_id, task_id = get_parent_id_and_task_id(latest_result_obj.dispatch_id)

            task_result = generate_task_result(
                task_id=task_id,
                status=Result.RUNNING,
            )

            send_task_update_to_dispatcher(parent_dispatch_id, task_result)

        return latest_result_obj

    elif task_execution_results["status"] == Result.FAILED:
        latest_result_obj._status = Result.FAILED

    elif task_execution_results["status"] == Result.CANCELLED:
        latest_result_obj._status = Result.CANCELLED

    elif task_execution_results["status"] == Result.COMPLETED:

        val = tasks_queue.get()

        print(f"Dispatch id: {dispatch_id}, tasks queue ALL CAPS: {val}", file=sys.stderr)

        tasks_queue.put(val)

        # If workflow is completed, post-process result
        if (
            not are_tasks_running(result_obj=latest_result_obj)
            and latest_result_obj.status is not Result.CANCELLED
        ):
            update_completed_workflow(latest_result_obj)

        elif not is_empty(tasks_queue) and latest_result_obj.status is not Result.CANCELLED:
            update_completed_tasks(dispatch_id, tasks_queue, latest_result_obj)

        # Finally if above two functions modified the tasks_queue or the result object
        if (
            is_empty(tasks_queue)
            and not are_tasks_running(result_obj=latest_result_obj)
            and latest_result_obj.status is not Result.CANCELLED
        ):
            update_completed_workflow(latest_result_obj)

    else:
        print(
            f"None of the above with status {task_execution_results['status']} and {is_empty(tasks_queue)}"
        )

    print(
        f"task id: {task_execution_results['node_id']} in dispatch id: {dispatch_id} marked as: {task_execution_results['status']}"
    )

    print(f"Result Object as :\n {latest_result_obj}")

    latest_result_obj = update_workflow_endtime(latest_result_obj)

    return latest_result_obj


def update_workflow_endtime(result_obj: Result) -> Result:
    """Update workflow end time if it has stopped running."""

    if result_obj.status not in [Result.RUNNING, Result.NEW_OBJ]:
        result_obj._end_time = datetime.now(timezone.utc)

    return result_obj


def update_completed_tasks(dispatch_id: str, tasks_queue: MPQ, result_obj: Result) -> Result:
    """Update completed tasks while the parent workflow is still not completed."""

    tasks_order_lod = tasks_queue.get()

    print(
        f"In update_completed_tasks with dispatch_id: {dispatch_id} and tasks_queue: {tasks_order_lod}"
    )

    tasks_dict = tasks_order_lod.pop(0)
    new_dispatch_id, new_tasks_order = zip(*tasks_dict.items())

    new_dispatch_id = new_dispatch_id[0]
    new_tasks_order = new_tasks_order[0]

    if tasks_order_lod:
        tasks_queue.put(tasks_order_lod)
    else:
        # Mark queue as empty
        # Removing tasks_dict from the queue so that the
        # new tasks order which will be sent in the case of
        # same dispatch_ids is an updated one and is
        # not confused with what's in the tasks queue

        # This will also mean that at this moment, tasks_queue
        # does not contain the order in which this last lattice's
        # tasks are going to be executed, which is to maintain uniformity
        # in the conditions in which dispatch_runnable_tasks is called.
        tasks_queue.put(None)

    if new_dispatch_id != dispatch_id:

        next_result_obj = get_result_object_from_result_service(dispatch_id=new_dispatch_id)

        dispatch_runnable_tasks(
            result_obj=next_result_obj, tasks_queue=tasks_queue, task_order=new_tasks_order
        )
    else:

        dispatch_runnable_tasks(
            result_obj=result_obj, tasks_queue=tasks_queue, task_order=new_tasks_order
        )

    return result_obj


def update_completed_workflow(result_obj: Result) -> Result:
    """Update the result for a completed lattice / sublattice workflow."""

    result_obj._result = _post_process(
        lattice=result_obj.lattice,
        task_outputs=result_obj.get_all_node_outputs(),
    )

    result_obj._status = Result.COMPLETED

    print(
        f"Dispatch id: {result_obj.dispatch_id}, REACHED UPDATE COMPLETED WORKFLOW",
        file=sys.stderr,
    )

    if is_sublattice_dispatch_id(result_obj.dispatch_id):
        print(
            f"Dispatch id: {result_obj.dispatch_id}, IS SUBLATTICE",
            file=sys.stderr,
        )

        parent_dispatch_id, task_id = get_parent_id_and_task_id(result_obj.dispatch_id)

        task_result = generate_task_result(
            task_id=task_id,
            end_time=datetime.now(timezone.utc),
            status=Result.COMPLETED,
            output=result_obj.result,
        )

        send_task_update_to_dispatcher(parent_dispatch_id, task_result)

    return result_obj



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
import os
import re
from typing import Optional, Tuple, Union

import sqlalchemy
from app.core.config import HOME_PATH, settings
from sqlalchemy import Column, DateTime, Integer, MetaData, String, Table
from sqlalchemy_utils import create_database, database_exists


class Database:
    def __init__(
        self,
        database_backend: str = None,
        db_name: str = None,
        db_endpoint: str = None,
        db_port: int = None,
        db_user: str = None,
        db_password: str = None,
    ):

        self.database_backend = database_backend or settings.DATABASE_BACKEND
        self.db_name = db_name or settings.DISPATCH_DB_NAME

        self.logger = logging.getLogger(__name__)

        if self.database_backend == "mysql":
            self.db_endpoint = db_endpoint or settings.DISPATCH_DB_ENDPOINT
            self.db_port = db_port or settings.DISPATCH_DB_PORT
            self.db_user = db_user or settings.DISPATCH_DB_USER

        self.connect(db_password or settings.DISPATCH_DB_PASSWORD)

        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        if not sqlalchemy.inspect(self.engine).has_table("workflow"):
            self.create_workflow_table()

    def connect(self, db_password: str = None):
        if self.database_backend == "sqlite":
            abs_db_path = os.path.join(HOME_PATH, "covalent", self.db_name)
            self.engine = sqlalchemy.create_engine(f"sqlite+pysqlite:///{abs_db_path}")
            self.logger.info("Connected to local SQLite database")
        elif self.database_backend == "mysql":
            self.engine = sqlalchemy.create_engine(
                f"mysql+pymysql://{self.db_user}:{db_password}@{self.db_endpoint}:{self.db_port}/{self.db_name}",
                pool_recycle=3600,
            )
            self.logger.info(f"Connected to MySQL database at {self.db_endpoint}")
        else:
            error_str = "Database backend not supported."
            self.logger.error(error_str)
            raise ValueError(error_str)

    def create_workflow_table(self):
        # TODO: Can metadata also be stored as a class variable?
        metadata = MetaData(self.engine)
        workflow_table = Table(
            "workflow",
            metadata,
            Column("id", String(256), primary_key=True),
            Column("results_filename", String(256)),
            Column("results_path", String(256)),
            # TODO: Modify the schema
        )

        metadata.create_all(self.engine)

        self.logger.info("Created workflow table")

    def create_task_table(self, table_name: str):
        name_pattern = re.compile("^[a-zA-Z0-9_]+$")
        if not re.match(name_pattern, table_name):
            raise ValueError("Invalid table name.")

        metadata = MetaData(self.engine)
        task_table = Table(
            table_name,
            metadata,
            Column("id", Integer, primary_key=True),
            # TODO: Decide on the schema later
        )

        metadata.create_all(self.engine)

        self.logger.info(f"Created task table {table_name}")

    def value(self, sql: str) -> str:
        self.logger.info("Executing SQL command.")
        self.logger.info(sql)

        connection = self.engine.connect()
        trans = connection.begin()
        response = connection.execute(sql)
        trans.commit()

        try:
            res = [r for r in response]
        except sqlalchemy.exc.ResourceClosedError:
            res = response.rowcount

        return res



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


from pathlib import Path
from typing import Any, List

import boto3
from app.core.config import settings
from app.schemas.common import HTTPExceptionSchema
from app.schemas.fs import DeleteResponse, UploadResponse
from fastapi import APIRouter, File, HTTPException, Query, UploadFile
from fastapi.responses import StreamingResponse
from minio import Minio

from ....core.localstoragebackend import LocalStorageBackend
from ....core.miniostoragebackend import MinioStorageBackend
from ....core.s3storagebackend import S3StorageBackend

router = APIRouter()

MINIO_USE_TLS = settings.MINIO_DISABLE_TLS.lower() not in ("true", "1")

minio_client = None
# TODO: support additional minio config options
if settings.FS_STORAGE_BACKEND == "minio":
    minio_client = Minio(
        endpoint=settings.MINIO_ENDPOINT,
        access_key=settings.MINIO_ACCESS_KEY,
        secret_key=settings.MINIO_SECRET_KEY,
        secure=MINIO_USE_TLS,
    )
    backend = MinioStorageBackend(minio_client, settings.FS_STORAGE_BUCKET)
elif settings.FS_STORAGE_BACKEND == "s3":
    # Reads AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, and AWS_DEFAULT_REGION from environment vars
    backend = S3StorageBackend(settings.FS_STORAGE_BUCKET)
else:
    backend = LocalStorageBackend(Path(settings.FS_LOCAL_STORAGE_ROOT), settings.FS_STORAGE_BUCKET)


@router.post("/upload", status_code=200, response_model=UploadResponse)
def upload_file(*, file: UploadFile, overwrite: bool = False) -> Any:
    """
    Upload a file
    """

    file.file.seek(0, 2)
    length = file.file.tell()
    file.file.seek(0)

    path, filename = backend.put(
        file.file, backend.bucket_name, file.filename, length, overwrite=overwrite
    )

    return {
        "filename": filename,
        "path": path,
    }


@router.get(
    "/download",
    status_code=200,
    response_class=StreamingResponse,
    responses={
        404: {"model": HTTPExceptionSchema, "description": "File was not found"},
        200: {
            "content": {"application/octet-stream": {}},
            "description": "Return binary content of file.",
        },
    },
)
def download_file(*, file_location: str) -> Any:
    """
    Download a file
    """
    stream = backend.get(backend.bucket_name, file_location)
    if not stream:
        raise HTTPException(status_code=404, detail="File not found")

    return StreamingResponse(
        stream,
        media_type="application/octet-stream",
        headers={"Content-Disposition": f'attachment; filename="{file_location}"'},
    )


@router.delete("/delete", status_code=200, response_model=DeleteResponse)
def delete_file(*, obj_names: List[str] = Query([])) -> DeleteResponse:
    deleted, failed = backend.delete(backend.bucket_name, obj_names)
    return {
        "deleted": deleted,
        "failed": failed,
    }



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
import os
import uuid
from io import BytesIO

import cloudpickle as pickle
from app.core.api import ResultsService
from app.core.config import settings
from app.core.queue import Queue
from app.schemas.submit import SubmitResponse
from fastapi import APIRouter, File, HTTPException

MQ_QUEUE_NAME = settings.MQ_QUEUE_NAME

router = APIRouter()


@router.post("/dispatch", status_code=202, response_model=SubmitResponse)
async def submit_workflow(*, result_pkl_file: bytes = File(...)) -> SubmitResponse:
    """
    Note: The object that contains the workflow function, interface to
    update attributes in the transport graph, inputs of the workflow,
    metadata, etc. is the result object that's why its type is ResultPickle
    which contains the pickled result object. Its use however varies in
    different components. We call it the result object because it is supposed
    to be the ultimate thing the user will get and will contain everything in the workflow.
    """

    queue = Queue(queue_name=MQ_QUEUE_NAME)
    results_svc = ResultsService()

    try:

        dispatch_id = str(uuid.uuid4())
        result_obj = pickle.loads(result_pkl_file)

        result_obj._dispatch_id = dispatch_id

        result_pkl_file = BytesIO(pickle.dumps(result_obj))

        await results_svc.create_result(result_pkl_file)

        await queue.publish(message_body_dict={"dispatch_id": dispatch_id})

        return {"dispatch_id": dispatch_id}

    except Exception as err:
        error_message = f"Error dispatching workflow due to error: {type(err)} {err}"
        logging.error(error_message)
        raise HTTPException(status_code=400, detail=error_message) from err



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


from multiprocessing import Queue as MPQ
from typing import Any, List

import cloudpickle as pickle
from app.core.cancel_workflow import cancel_workflow_execution
from app.core.dispatch_workflow import dispatch_workflow
from app.core.dispatcher_logger import logger
from app.core.update_workflow import update_workflow_results
from app.core.utils import (
    get_result_object_from_result_service,
    is_empty,
    is_sublattice_dispatch_id,
    update_result_and_ui,
)
from app.schemas.workflow import (
    BatchCancelWorkflowResponse,
    CancelWorkflowResponse,
    DispatchWorkflowResponse,
    UpdateWorkflowResponse,
)
from fastapi import APIRouter, File, Query

from covalent._results_manager import Result

workflow_tasks_queue = MPQ()
workflow_status_queue = MPQ()

# Using sentinel to indicate that the queue is empty since MPQ.empty() is an unreliable method
workflow_tasks_queue.put(None)
workflow_status_queue.put(None)


router = APIRouter()

# Do we need this here?
mock_result = {
    "status": "COMPLETED",
    "result": {
        "dispatch_id": "",
        "results_dir": "",
        "status": "COMPLETED",
        "graph": {
            "links": [{"edge_name": "data", "param_type": "kwarg", "source": 0, "target": 1}],
            "nodes": [
                {
                    "name": "load_data",
                    "metadata": {"executor": "<LocalExecutor>"},
                    "function_string": "@ct.electron\ndef load_data():\n    iris = datasets.load_iris()\n    perm = permutation(iris.target.size)\n    iris.data = iris.data[perm]\n    iris.target = iris.target[perm]\n    return iris.data, iris.target\n\n\n",
                    "start_time": "2022-02-28T23:21:17.844268+00:00",
                    "end_time": "2022-02-28T23:21:17.853741+00:00",
                    "status": "COMPLETED",
                    "output": ["[[5.  3.4 1.5 0.2]]"],
                    "error": None,
                    "sublattice_result": None,
                    "stdout": "",
                    "stderr": "",
                    "id": 0,
                    "doc": None,
                    "kwargs": None,
                }
            ],
        },
    },
}


logger.warning("Dispatcher Service Started")


@router.post("/{dispatch_id}", status_code=202, response_model=DispatchWorkflowResponse)
def submit_workflow(*, dispatch_id: str) -> Any:
    """
    Submit a workflow
    """

    logger.warning(f"Inside submit_workflow with dispatch_id: {dispatch_id}")

    # Change workflow status to RUNNING
    workflow_status_queue.get()

    workflow_status_queue.put(Result.RUNNING)

    # Get the result object
    result_obj = get_result_object_from_result_service(dispatch_id=dispatch_id)

    # Dispatch the workflow
    dispatch_workflow(result_obj=result_obj, tasks_queue=workflow_tasks_queue)

    vv = workflow_tasks_queue.get()

    logger.warning(f"IT SHOULD NOT BE EMPTY HERE {vv}")

    workflow_tasks_queue.put(vv)

    logger.warning(f"Inside submit_workflow dispatching done with dispatch_id: {dispatch_id}")

    return {"response": f"{dispatch_id} workflow dispatched successfully"}


@router.delete("/{dispatch_id}", status_code=200, response_model=CancelWorkflowResponse)
def cancel_workflow(*, dispatch_id: str) -> CancelWorkflowResponse:
    """
    Cancel a workflow
    """

    result_obj = get_result_object_from_result_service(dispatch_id=dispatch_id)

    success = cancel_workflow_execution(result_obj)

    # Note - The queue should be populated in theory.
    if not is_empty(workflow_tasks_queue):
        workflow_tasks_queue.get()
        workflow_tasks_queue.put(None)

    # Empty queue when workflow is terminated
    if success:
        workflow_status_queue.get()
        workflow_status_queue.put(None)

        return {"response": f"{dispatch_id} workflow cancelled successfully"}
    else:
        return {"response": f"{dispatch_id} workflow did not cancel successfully"}


@router.delete("/cancel", status_code=200, response_model=BatchCancelWorkflowResponse)
def cancel_workflows(*, dispatch_ids: List[str] = Query([])) -> BatchCancelWorkflowResponse:
    """
    Cancel a set of workflows
    """

    # Mock response here
    mock_response = {
        "cancelled": ["string"],
        "failed": ["string"],
    }

    return mock_response


@router.put("/{dispatch_id}", status_code=200, response_model=UpdateWorkflowResponse)
def update_workflow(
    *, dispatch_id: str, task_execution_results: bytes = File(...)
) -> UpdateWorkflowResponse:
    """
    Update a workflow
    """

    logger.warning(f"Received to update task for id: {dispatch_id}")

    task_execution_results = pickle.loads(task_execution_results)
    task_id = task_execution_results["task_id"]
    del task_execution_results["task_id"]
    task_execution_results["node_id"] = task_id

    updated_result_obj = update_workflow_results(
        task_execution_results=task_execution_results,
        dispatch_id=dispatch_id,
        tasks_queue=workflow_tasks_queue,
    )

    # Empty queue when workflow is no longer running (completed # or failed)
    if updated_result_obj.status != Result.RUNNING and not is_sublattice_dispatch_id(
        updated_result_obj.dispatch_id
    ):

        logger.warning("updated_result_obj status is not RUNNING")

        workflow_status_queue.get()
        workflow_status_queue.put(None)

    update_result_and_ui(updated_result_obj, task_id)

    logger.warning("updated_result_obj status is done")

    return {"response": f"{dispatch_id} workflow updated successfully"}


@router.get("/status", status_code=200)
def get_status():

    if is_empty(workflow_status_queue):
        return {"status": "EMPTY"}

    status = workflow_status_queue.get()
    workflow_status_queue.put(status)

    return {"status": f"{status}"}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Utility functions for the Dispatcher microservice."""

import os
import sys
from datetime import datetime, timezone
from io import BytesIO
from multiprocessing import Queue as MPQ
from queue import Empty
from typing import Any, Dict, List, Tuple

import cloudpickle as pickle
import requests
from app.core.dispatcher_logger import logger
from app.core.get_svc_uri import DispatcherURI, ResultsURI, RunnerURI, UIBackendURI
from dotenv import load_dotenv

from covalent._results_manager import Result
from covalent._shared_files.context_managers import active_lattice_manager
from covalent._shared_files.defaults import (
    attr_prefix,
    electron_dict_prefix,
    electron_list_prefix,
    generator_prefix,
    parameter_prefix,
    prefix_separator,
    sublattice_prefix,
    subscript_prefix,
)
from covalent._workflow.lattice import Lattice

load_dotenv()


BASE_URI = os.environ.get("BASE_URI")


def is_empty(mp_queue: MPQ):
    """The underlying assumption is that mpq contains only one element at any given time."""

    try:
        elem = mp_queue.get(timeout=1)
    except Empty:
        mp_queue.put(None)
        return True

    status = elem is None
    mp_queue.put(elem)
    return status


def preprocess_transport_graph(task_id: int, task_name: str, result_obj: Result) -> Result:
    """Ensure that the execution status of the task nodes in the transport graph are initialized
    properly."""

    is_executable_node = True

    if task_name.startswith((subscript_prefix, generator_prefix, parameter_prefix, attr_prefix)):
        if task_name.startswith(parameter_prefix):
            output = result_obj.transport_graph.get_node_value(task_id, "value")

        else:
            parent = result_obj.transport_graph.get_dependencies(task_id)[0]
            output = result_obj._get_node_output(parent)

            if task_name.startswith(attr_prefix):
                attr = result_obj.transport_graph.get_node_value(task_id, "attribute_name")
                output = getattr(output, attr)
            else:
                key = result_obj.transport_graph.get_node_value(task_id, "key")

                if output is None:
                    print(f"Failed task name: {task_name}", file=sys.stderr)
                output = output[key]

        result_obj._update_node(
            node_id=task_id,
            node_name=f"{task_name}({task_id})",
            start_time=datetime.now(timezone.utc),
            end_time=datetime.now(timezone.utc),
            status=Result.COMPLETED,
            output=output,
        )

        is_executable_node = False

    return result_obj, is_executable_node


def _post_process(lattice: Lattice, task_outputs: Dict) -> Any:
    """
    Post processing function to be called after the lattice execution.
    This takes care of executing statements that were not an electron
    but were inside the lattice's function. It also replaces any calls
    to an electron with the result of that electron execution, hence
    preventing a local execution of electron's function.

    Note: Here `node_outputs` is used instead of `electron_outputs`
    since an electron can be called multiple times with possibly different
    arguments, but every time it's called, it will be executed as a separate node.
    Thus, output of every node is used.

    Args:
        lattice: Lattice object that was dispatched.
        task_outputs: Dictionary containing the output of all the tasks.

    Returns:
        result: The result of the lattice function.
    """

    ordered_task_outputs = [
        val
        for key, val in task_outputs.items()
        if not key.startswith(prefix_separator) or key.startswith(sublattice_prefix)
    ]

    with active_lattice_manager.claim(lattice):
        lattice.post_processing = True
        lattice.electron_outputs = ordered_task_outputs
        result = lattice.workflow_function(*lattice.args, **lattice.kwargs)
        lattice.post_processing = False
        return result


def get_task_inputs(task_id: int, node_name: str, result_obj: Result) -> Dict:
    """
    Return the required inputs for a task execution.
    This makes sure that any node with child nodes isn't executed twice and fetches the
    result of parent node to use as input for the child node.

    Args:
        task_id: Node id of this task in the transport graph.
        node_name: Name of the node.
        result_obj: Result object to be used to update and store execution related
                       info including the results.

    Returns:
        inputs: Input dictionary to be passed to the task containing args, kwargs,
                and any parent node execution results if present.
    """

    if node_name.startswith(electron_list_prefix):
        values = [
            result_obj._get_node_output(parent)
            for parent in result_obj.transport_graph.get_dependencies(task_id)
        ]
        task_inputs = {"args": [], "kwargs": {"x": values}}
    elif node_name.startswith(electron_dict_prefix):
        values = {}
        for parent in result_obj.transport_graph.get_dependencies(task_id):
            key = result_obj.transport_graph.get_edge_value(parent, task_id, "edge_name")
            value = result_obj._get_node_output(parent)
            values[key] = value
        task_inputs = {"args": [], "kwargs": {"x": values}}
    else:
        task_inputs = {"args": [], "kwargs": {}}

        for parent in result_obj.transport_graph.get_dependencies(task_id):

            param_type = result_obj.transport_graph.get_edge_value(parent, task_id, "param_type")

            value = result_obj._get_node_output(parent)

            if param_type == "arg":
                task_inputs["args"].append(value)

            elif param_type == "kwarg":
                key = result_obj.transport_graph.get_edge_value(parent, task_id, "edge_name")

                task_inputs["kwargs"][key] = value
    return task_inputs


def is_sublattice(task_name: str = None) -> bool:
    """Check if the transport graph task node is a sublattice. When the workflow is first
    dispatched, the build graph step is responsible for attaching a `sublattice` prefix to the
    corresponding task node."""

    return task_name.startswith(sublattice_prefix)


def are_tasks_running(result_obj: Result) -> bool:
    """Check if any of the tasks are still running. A task is considered not `running` if it was completed, failed or cancelled."""

    return any(
        result_obj._get_node_status(task_id) in [Result.RUNNING, Result.NEW_OBJ]
        for task_id in range(result_obj._num_nodes)
        if not result_obj._get_node_name(task_id).startswith(parameter_prefix)
    )


def get_task_order(result_obj: Result) -> List[List]:
    """Find the order in which the tasks need to be executed. At the current moment this is
    based simply on the topologically sorted task nodes in the graph. In the future,
    this function can become much more sophisticated and optimized.
    """

    return result_obj.transport_graph.get_topologically_sorted_graph()


def send_task_list_to_runner(dispatch_id, tasks_list) -> List[int]:

    logger.warning(f"Inside send_task_list_to_runner with dispatch_id {dispatch_id}")
    logger.warning(f"Inside send_task_list_to_runner with tasks_list {tasks_list}")

    # Example tasks_list:
    # tasks_list = [
    #     {
    #         "task_id": 0,
    #         "func": result_object.transport_graph.get_node_value(0, "function"),
    #         "args": [2 + 2],
    #         "kwargs": {},
    #         "executor": result_object.transport_graph.get_node_value(0, "metadata")[
    #             "executor"
    #         ],
    #         "results_dir": result_object.results_dir,
    #     },
    #     {
    #         "task_id": 2,
    #         "func": result_object.transport_graph.get_node_value(2, "function"),
    #         "args": [2, 10],
    #         "kwargs": {},
    #         "executor": result_object.transport_graph.get_node_value(2, "metadata")[
    #             "executor"
    #         ],
    #         "results_dir": result_object.results_dir,
    #     },
    # ]
    # response = requests.post(url=url_endpoint, files={"tasks": pickle.dumps(tasks_list)})

    # Set the url endpoint
    url_endpoint = RunnerURI().get_route(f"workflow/{dispatch_id}/tasks")

    # Send the tasks list as file
    response = requests.post(url=url_endpoint, files={"tasks": BytesIO(pickle.dumps(tasks_list))})

    # Raise error if occurred
    response.raise_for_status()

    return response.json()["left_out_task_ids"]


def send_result_object_to_result_service(result_object: Result):
    """Send result object to the result microservice."""

    url_endpoint = ResultsURI().get_route("workflow/results")

    response = requests.post(
        url=url_endpoint, files={"result_pkl_file": BytesIO(pickle.dumps(result_object))}
    )
    response.raise_for_status()

    return response.text


def send_task_update_to_result_service(dispatch_id: str, task_execution_result: dict):

    url_endpoint = ResultsURI().get_route(f"workflow/results/{dispatch_id}")

    response = requests.put(
        url=url_endpoint, files={"task": BytesIO(pickle.dumps(task_execution_result))}
    )
    response.raise_for_status()

    return response.text


def send_task_update_to_ui(dispatch_id: str, task_id: int):
    """Send task update to UI backend microservice."""

    url_endpoint = UIBackendURI().get_route(f"ui/workflow/{dispatch_id}/task/{task_id}")

    response = requests.put(url=url_endpoint)
    response.raise_for_status()

    return response.text


def get_result_object_from_result_service(dispatch_id: str):
    logger.warning(f"getting result object with id {dispatch_id}")

    url_endpoint = ResultsURI().get_route(f"workflow/results/{dispatch_id}")

    response = requests.get(url=url_endpoint, stream=True)
    response.raise_for_status()

    return pickle.loads(response.content)


def update_result_and_ui(result_obj: Result, task_id: int) -> Dict[str, str]:
    """Write the updated result to the database and update the UI."""

    resp_1 = send_result_object_to_result_service(result_obj)
    resp_2 = send_task_update_to_ui(dispatch_id=result_obj.dispatch_id, task_id=task_id)
    return {"update_result_response": resp_1, "update_ui_response": resp_2}


def send_cancel_task_to_runner(dispatch_id: str, task_id: int) -> Tuple[str, str]:

    url_endpoint = RunnerURI().get_route(f"workflow/{dispatch_id}/task/{task_id}")
    response = requests.delete(url=url_endpoint)
    response.raise_for_status()

    return response.json()["cancelled_dispatch_id"], response.json()["cancelled_task_id"]


def is_sublattice_dispatch_id(dispatch_id: str):
    return ":" in dispatch_id


def send_task_update_to_dispatcher(dispatch_id, task_result):

    url = DispatcherURI().get_route(f"workflow/{dispatch_id}")

    logger.warning(
        f"Sending task result to get updated in dispatcher with task result: {task_result}"
    )
    logger.warning(f"URL: {url}")

    response = requests.put(url=url, files={"task_execution_results": pickle.dumps(task_result)})

    logger.warning("Put done with response")

    response.raise_for_status()


def generate_task_result(
    task_id,
    start_time=None,
    end_time=None,
    status=None,
    output=None,
    error=None,
    stdout=None,
    stderr=None,
    info=None,
):

    return {
        "task_id": task_id,
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "output": output,
        "error": error,
        "stdout": stdout,
        "stderr": stderr,
        "info": info,
    }


def get_parent_id_and_task_id(dispatch_id: str):

    if not is_sublattice_dispatch_id(dispatch_id):
        return False, False

    splits = dispatch_id.split(":")
    parent_dispatch_id, task_id = ":".join(splits[:-1]), splits[-1]
    return parent_dispatch_id, int(task_id)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Sequence

from pydantic import BaseModel


class BaseNode(BaseModel):
    name: str
    start_time: datetime
    end_time: datetime
    status: str
    output: Any
    error: Optional[str]
    stdout: str
    stderr: str


class Link(BaseModel):
    edge_name: str
    param_type: str
    source: int
    target: int


class Node(BaseNode):
    id: int


class Graph(BaseModel):
    nodes: Sequence[Node]
    links: Sequence[Link]


class Result(BaseModel):
    dispatch_id: str
    results_dir: str
    status: str
    graph: Graph


class InsertResultResponse(BaseModel):
    dispatch_id: str


class UpdateResultResponse(BaseModel):
    response: str


class ResultPickle(BaseModel):
    result_object: bytes


class ResultFormats(str, Enum):
    BINARY = "binary"
    JSON = "json"


class DeleteResultResponse(BaseModel):
    deleted: int
    failed: int



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import os
from typing import List, Union

from dotenv import find_dotenv, load_dotenv
from pydantic import AnyHttpUrl, BaseSettings, validator

HOME_PATH = os.environ.get("XDG_CACHE_HOME") or (os.environ["HOME"] + "/.cache")
load_dotenv(os.path.join(HOME_PATH, "covalent", ".env"))


class Settings(BaseSettings):
    API_V0_STR: str = "/api/v0"
    BACKEND_CORS_ORIGINS: List[AnyHttpUrl] = []
    RUNNER_SVC_PORT: int = 8003
    RUNNER_SVC_HOST: str = "localhost"
    RESULTS_SVC_PORT: int = 8005
    RESULTS_SVC_HOST: str = "localhost"
    DISPATCHER_SVC_PORT: int = 8002
    DISPATCHER_SVC_HOST: str = "localhost"

    @validator("BACKEND_CORS_ORIGINS", pre=True)
    def assemble_cors_origins(cls, v: Union[str, List[str]]) -> Union[List[str], str]:
        if isinstance(v, str) and not v.startswith("["):
            return [i.strip() for i in v.split(",")]
        elif isinstance(v, (list, str)):
            return v
        raise ValueError(v)

    class Config:
        case_sensitive = True


settings = Settings()

print("Configuration:")
print(settings)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

from copy import deepcopy
from functools import wraps
from io import BytesIO
from typing import Callable, List, Union

import cloudpickle as pickle
import requests

from .._results_manager.result import Result
from .._workflow.lattice import Lattice
from . import get_svc_uri


def dispatch(
    orig_lattice: Lattice,
    queuer_addr: str = get_svc_uri.QueuerURI().get_route("submit/dispatch"),
) -> Callable:
    """
    Wrapping the dispatching functionality to allow input passing
    and server address specification.

    Afterwards, send the lattice to the dispatcher server and return
    the assigned dispatch id.

    Args:
        orig_lattice: The lattice/workflow to send to the dispatcher server.
        dispatcher_addr: The address of the dispatcher server.

    Returns:
        Wrapper function which takes the inputs of the workflow as arguments
    """

    @wraps(orig_lattice)
    def wrapper(*args, **kwargs) -> str:
        """
        Send the lattice to the dispatcher server and return
        the assigned dispatch id.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The dispatch id of the workflow.
        """

        lattice = deepcopy(orig_lattice)

        lattice.build_graph(*args, **kwargs)

        pickled_res = pickle.dumps(Result(lattice, lattice.metadata["results_dir"]))

        r = requests.post(queuer_addr, files={"result_pkl_file": BytesIO(pickled_res)})
        r.raise_for_status()

        # Returns assigned dispatch id
        return r.json()["dispatch_id"]

    return wrapper


def dispatch_sync(
    lattice: Lattice,
    queuer_addr: str = get_svc_uri.QueuerURI().get_route("submit/dispatch"),
) -> Callable:
    """
    Wrapping the synchronous dispatching functionality to allow input
    passing and server address specification.

    Afterwards, sends the lattice to the dispatcher server and return
    the result of the executed workflow.

    Args:
        orig_lattice: The lattice/workflow to send to the dispatcher server.
        dispatcher_addr: The address of the dispatcher server.

    Returns:
        Wrapper function which takes the inputs of the workflow as arguments
    """

    @wraps(lattice)
    def wrapper(*args, **kwargs) -> Result:
        """
        Send the lattice to the dispatcher server and return
        the result of the executed workflow.

        Args:
            *args: The inputs of the workflow.
            **kwargs: The keyword arguments of the workflow.

        Returns:
            The result of the executed workflow.
        """

        return get_result(
            dispatch_id=dispatch(lattice, queuer_addr)(*args, **kwargs),
            wait=True,
        )

    return wrapper


def _retrieve_result_response(session: requests.Session, dispatch_id: str) -> requests.Response:

    response = session.get(
        get_svc_uri.ResultsURI().get_route(f"workflow/results/{dispatch_id}"), stream=True
    )

    response.raise_for_status()

    return response


def _poll_result(
    session: requests.Session, dispatch_id: str, wait: bool = False
) -> requests.Response:

    response = _retrieve_result_response(session, dispatch_id)

    if wait:
        result_object: Result = pickle.loads(response.content)

        while result_object.status not in [Result.COMPLETED, Result.FAILED, Result.CANCELLED]:
            response = _retrieve_result_response(session, dispatch_id)
            result_object: Result = pickle.loads(response.content)

    return response


def get_result(dispatch_id: str, download=False, wait=False):

    session = requests.Session()

    response = _poll_result(session, dispatch_id, wait)

    if not download:
        return pickle.loads(response.content)

    filename = f"result_{dispatch_id}.pkl"
    with open(filename, "wb") as f:
        f.write(response.content)

    return filename


def cancel_workflow(dispatch_id: str):
    response = requests.delete(get_svc_uri.DispatcherURI().get_route(f"workflow/{dispatch_id}"))

    response.raise_for_status()

    return response.json()


def sync(dispatch_id: Union[List[str], str]) -> None:

    session = requests.Session()

    workflow_list = dispatch_id if isinstance(dispatch_id, list) else [dispatch_id]

    for workflow in workflow_list:
        _poll_result(session, workflow, True)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
from abc import ABC
from http import client
from pathlib import Path
from typing import BinaryIO, Generator, List, Union

import boto3
import botocore.exceptions
from fastapi import HTTPException

from .storagebackend import StorageBackend

logger = logging.getLogger(__name__)


class S3StorageBackend(ABC):
    """Thin wrapper for S3 client

    Attributes:
        client: S3 client instance
        bucket_name: current working bucket name (default: "default")
    """

    def __init__(self, bucket_name="default"):
        self.bucket_name = bucket_name
        self.client = boto3.client("s3")

    def handle_client_error(self, e):
        err_code = e.response["Error"]["Code"]
        logger.warn(f"S3 Client Error: {err_code}")
        if err_code == "AccessDenied":
            raise HTTPException(
                401,
                detail="S3 client error: AccessDenied (Likely incorrect AWS credentials or configuration).",
            )
        else:
            raise HTTPException(400, detail=f"S3 client error: {err_code}")

    def get(self, bucket_name: str, object_name: str) -> Union[Generator[bytes, None, None], None]:
        """Get object from storage.

        Args:
            bucket_name: name of the bucket
            object_name: name of the object

        Returns:
            A generator yielding a byte stream or None if an
            error occurred in retrieving the object

        """

        try:
            resp = self.client.get_object(Bucket=bucket_name, Key=object_name)
        except (botocore.exceptions.ClientError) as e:
            # TODO: better logging
            logger.warn(f"Exception in S3 client when fetching object: {object_name}")
            self.handle_client_error(e)
            return None

        return resp["Body"]

    def put(
        self,
        data: BinaryIO,
        bucket_name: str,
        object_name: str,
        length: int,
        metadata: dict = None,
        overwrite: bool = False,
    ) -> (str, str):

        """Upload object to storage.

        Args:
            data: a binary file-like object
            bucket_name: name of the bucket
            object_name: name of the destination object
            length: number of bytes to upload
        Returns:
            (bucket_name, object_name) if write succeeds and ("", "") otherwise

        """

        try:
            res = self.client.put_object(
                Bucket=bucket_name, Key=object_name, ContentLength=length, Body=data
            )
        except (botocore.exceptions.ClientError) as e:
            # TODO: better logging
            logger.warn(f"Exception in S3 client when uploading object: {object_name}")
            self.handle_client_error(e)
            res = None

        if res:
            return (f"s3://{bucket_name}/{object_name}", object_name)
        else:
            return ("", "")

    def delete(self, bucket_name: str, object_names: List[str]):
        deleted_objects = []
        failed = []
        try:
            res = self.client.delete_objects(
                Bucket=bucket_name,
                Delete={
                    "Objects": list(map(lambda obj_name: {"Key": obj_name}, object_names)),
                    "Quiet": False,
                },
            )

            try:
                deleted_objects = res["Deleted"]
            except KeyError:
                deleted_objects = []

            try:
                failed = res["Errors"]
            except KeyError:
                failed = []

        except (botocore.exceptions.ClientError) as e:
            failed = object_names
            self.handle_client_error(e)

        deleted_objects = list(map(lambda obj_metadata: obj_metadata["Key"], deleted_objects))
        failed = list(map(lambda obj_metadata: obj_metadata["Key"], failed))

        return deleted_objects, failed



from datetime import datetime
from typing import Any, Optional, Sequence

from pydantic import BaseModel


class BaseNode(BaseModel):
    name: str
    start_time: datetime
    end_time: datetime
    status: str
    output: Any
    error: Optional[str]
    stdout: str
    stderr: str


class Link(BaseModel):
    edge_name: str
    param_type: str
    source: int
    target: int


class Node(BaseNode):
    id: int


class Graph(BaseModel):
    nodes: Sequence[Node]
    links: Sequence[Link]


class Result(BaseModel):
    dispatch_id: str
    status: str
    graph: Graph



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
import os
from pathlib import Path

from app.api.api_v0.api import api_router
from app.api.health import health_router
from app.core.config import settings
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi_socketio import SocketManager
from starlette.exceptions import HTTPException

BASE_PATH = Path(__file__).resolve().parent
FRONTEND_PATH = "/webapp/build"


class SinglePageApp(StaticFiles):
    """
    Allow URL paths (e.g. /preview or /[dispatch_id]) to be handled by the web
    app by overriding the handling of 404 exceptions and passing control to the
    app instead (/index.html)
    """

    async def get_response(self, path: str, scope):
        try:
            response = await super().get_response(path, scope)
        except HTTPException as e:
            if e.status_code == 404:
                # return /index.html
                response = await super().get_response(".", scope)
            else:
                raise e
        return response


app = FastAPI(title="Covalent UI Backend Service API")

sio = SocketManager(app=app)

logging.basicConfig(level=logging.DEBUG)

app.include_router(api_router, prefix=settings.API_V0_STR)
app.include_router(health_router)

app.mount(
    "/",
    SinglePageApp(directory=f"{BASE_PATH}{FRONTEND_PATH}", html=True, check_dir=False),
    name="static",
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

if __name__ == "__main__":
    # Use this for debugging purposes only
    import uvicorn

    uvicorn.run(
        "main:app",
        host=settings.UI_SVC_HOST,
        port=settings.UI_SVC_PORT,
        log_level="debug",
        reload=False,
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

"""Workflow cancel functionality."""

import sys
from datetime import datetime, timezone
from typing import List, Tuple

from app.core.get_svc_uri import RunnerURI
from app.core.utils import (
    generate_task_result,
    get_parent_id_and_task_id,
    get_result_object_from_result_service,
    is_sublattice,
    is_sublattice_dispatch_id,
    send_cancel_task_to_runner,
    send_result_object_to_result_service,
    send_task_update_to_dispatcher,
)

from covalent._results_manager import Result


def cancel_workflow_execution(
    result_obj: Result, task_id_batch: List[Tuple[str, int]] = None
) -> bool:
    """Main cancel function. Called by the user via ct.cancel(dispatch_id). The task_id_batch is composed of both
    dispatch id and task ids in the form of a tuple."""

    workflow_cancelled = True

    (
        dispatch_and_task_id_lot,
        unrun_tasks_lot,
    ) = task_id_batch or get_cancellable_dispatch_and_task_ids(result_obj)

    # Using the cancellation_status variable because we still want to continue
    # attempting cancellation of further tasks even if one of them has failed
    for dispatch_id, task_id in dispatch_and_task_id_lot:
        task_cancelled = cancel_task(dispatch_id, task_id)

        if is_sublattice_dispatch_id(dispatch_id) and task_cancelled:
            parent_dispatch_id, task_id = get_parent_id_and_task_id(dispatch_id)

            task_result = generate_task_result(
                task_id=task_id,
                end_time=datetime.now(timezone.utc),
                status=Result.CANCELLED,
            )

            send_task_update_to_dispatcher(parent_dispatch_id, task_result)

        if not task_cancelled:
            workflow_cancelled = False

    if workflow_cancelled and not task_id_batch:
        for dispatch_id, task_id in unrun_tasks_lot:
            cancel_unrun_tasks(dispatch_id, task_id)

    print(f"Cancelled workflow: {workflow_cancelled}", file=sys.stderr)

    return workflow_cancelled


def cancel_unrun_tasks(dispatch_id: str, task_id: int) -> None:

    task_result = generate_task_result(
        task_id=task_id,
        end_time=datetime.now(timezone.utc),
        status=Result.CANCELLED,
    )

    send_task_update_to_dispatcher(dispatch_id, task_result)


def cancel_task(dispatch_id: str, task_id: int) -> bool:
    """Asks the Runner API to cancel the execution of these tasks and returns the status of whether it was
    successful."""

    cancelled_dispatch_id, cancelled_task_id = send_cancel_task_to_runner(
        dispatch_id=dispatch_id, task_id=task_id
    )
    print(
        f"Cancel dispatch id: {type(cancelled_dispatch_id)}, cancel task id: {type(cancelled_task_id)}",
        file=sys.stderr,
    )
    if (
        (cancelled_dispatch_id and cancelled_task_id)
        and (cancelled_dispatch_id == dispatch_id)
        and (cancelled_task_id == str(task_id))
    ):
        return True

    return False


def get_cancellable_dispatch_and_task_ids(result_obj: Result) -> List[Tuple[str, int]]:
    """Get all the task ids and the corresponding dispatch ids for a given lattice. When a sublattice is encountered,
    the dispatch id corresponding to the sublattice `dispatch_id:task_id` is used."""

    dispatch_and_task_id_lot = []
    unrun_tasks_lot = []

    task_order = result_obj.transport_graph.get_topologically_sorted_graph()
    task_ids = [id for sublist in task_order for id in sublist]

    for task_id in task_ids:

        if result_obj._get_node_status(task_id) == Result.RUNNING:
            lot_to_update = dispatch_and_task_id_lot

        elif result_obj._get_node_status(task_id) == Result.NEW_OBJ:
            lot_to_update = unrun_tasks_lot

        else:
            continue

        task_name = result_obj._get_node_name(task_id)
        if not is_sublattice(task_name):
            lot_to_update.append((result_obj.dispatch_id, task_id))
        else:

            # Super lattice's sublattice task needs to get cancelled
            # too right here
            unrun_tasks_lot.append((result_obj.dispatch_id, task_id))

            sublattice_result_obj = get_result_object_from_result_service(
                f"{result_obj.dispatch_id}:{task_id}"
            )
            lot_1, lot_2 = get_cancellable_dispatch_and_task_ids(sublattice_result_obj)

            dispatch_and_task_id_lot = lot_1 + dispatch_and_task_id_lot
            unrun_tasks_lot = lot_2 + unrun_tasks_lot

    return dispatch_and_task_id_lot, unrun_tasks_lot



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.
import logging
import os
import sqlite3
from typing import Optional, Tuple, Union
from urllib.parse import urljoin

import requests
from app.core.config import settings
from app.core.get_svc_uri import DataURI
from fastapi import HTTPException


class APIService:
    def __init__(self, BASE_URI: str):
        if BASE_URI[len(BASE_URI) - 1] != "/":
            BASE_URI = f"{BASE_URI}/"
        self.BASE_URI = BASE_URI

    def _get_route(self, path: str):
        return urljoin(self.BASE_URI, path)

    def _format(self, response, content=False):
        if response.status_code >= 400:
            error_body = response.json()
            raise HTTPException(status_code=response.status_code, detail=error_body["detail"])
        if content:
            return response.content
        return response.json()

    def post(self, path, json={}, params={}, data={}, files={}):
        route = self._get_route(path)
        return self._format(requests.post(route, json=json, params=params, data=data, files=files))

    def get(self, path, params={}, stream=False, content=False):
        route = self._get_route(path)
        return self._format(requests.get(route, params=params, stream=stream), content)

    def delete(self, path, params={}):
        route = self._get_route(path)
        return self._format(requests.delete(route, params=params))

    def patch(self, path, json={}, params={}, data={}):
        route = self._get_route(path)
        return self._format(requests.patch(route, json=json, params=params, data=data))

    def put(self, path, json={}, params={}, data={}):
        route = self._get_route(path)
        return self._format(requests.put(route, json=json, params=params, data=data))


class DataService(APIService):
    def __init__(self):
        super().__init__(str(DataURI().get_base_url()))

    async def download(self, filename: str):
        return self.get(
            "/api/v0/fs/download", params={"file_location": filename}, stream=True, content=True
        )

    async def upload(self, result_pkl_file: bytes):
        return self.post(
            "/api/v0/fs/upload",
            files=[("file", ("result.pkl", result_pkl_file, "application/octet-stream"))],
        )



from app.core.config import settings
from furl import furl


class ServiceURI:
    def __init__(
        self, scheme: str = "http", host: str = "localhost", port=None, preffix="api/v0"
    ) -> None:
        self.scheme = scheme
        self.host = host
        self.port = port
        self.preffix = preffix

    def get_base_url(self):
        base_url = furl().set(scheme=self.scheme, host=self.host, port=self.port)
        if self.preffix:
            base_url.set(path=self.preffix)
        return base_url

    def get_route(self, path: str):
        base_url = self.get_base_url().copy()
        base_url.path /= path
        return base_url.url


class RunnerURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(port=settings.RUNNER_SVC_PORT, host=settings.RUNNER_SVC_HOST)


class ResultsURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(port=settings.RESULTS_SVC_PORT, host=settings.RESULTS_SVC_HOST)


class DispatcherURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(port=settings.DISPATCHER_SVC_PORT, host=settings.DISPATCHER_SVC_HOST)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.
import logging

from aiolimiter import AsyncLimiter
from app.schemas.ui import DrawRequest, UpdateUIResponse
from fastapi import APIRouter, BackgroundTasks, Request
from fastapi.encoders import jsonable_encoder

dispatch_set = set()
limiter = AsyncLimiter(1, 2)

router = APIRouter()


async def notify_frontend(app, topic: str, payload):
    await app.sio.emit(topic, payload)


async def throttle_request_update_notify(app, dispatch_id, task_id):
    global dispatch_set
    if (dispatch_id, task_id) not in dispatch_set:
        dispatch_set.add((dispatch_id, task_id))
        async with limiter:
            dispatch_set.remove((dispatch_id, task_id))
            logging.debug(
                f"Emitting websocket event to update task {task_id} in workflow {dispatch_id}"
            )
            await notify_frontend(
                app, "result-update", {"result": {"dispatch_id": dispatch_id, "task_id": task_id}}
            )


@router.put(
    "/workflow/{dispatch_id}/task/{task_id}", status_code=200, response_model=UpdateUIResponse
)
async def update_ui(
    *, dispatch_id: str, task_id: int, request: Request, background_tasks: BackgroundTasks
) -> UpdateUIResponse:

    """
    API Endpoint (/api/workflow/task) to update ui frontend
    """

    # throttle notify frontend calls in background
    background_tasks.add_task(throttle_request_update_notify, request.app, dispatch_id, task_id)

    return {"response": "UI Updated"}


@router.post("/workflow/draft", status_code=200, response_model=UpdateUIResponse)
async def draw_draft(*, payload: DrawRequest, request: Request) -> UpdateUIResponse:
    """
    API Endpoint (/api/workflow/draft) to draw workflow draft
    """

    await notify_frontend(request.app, "draw-request", jsonable_encoder(payload))

    return {"response": "UI Workflow Draft Sent"}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.
import os
from pathlib import Path

from fastapi import APIRouter, Response, status
from pydantic import BaseModel

UI_BUILD_PATH = Path(__file__).parents[2].resolve() / Path("webapp/build")

health_router = APIRouter()


class HealthResponse(BaseModel):
    is_healthy: bool = True


class ReadyResponse(BaseModel):
    is_ready: bool = True
    is_dashboard_available: bool = True


@health_router.get("/healthz", status_code=200, response_model=HealthResponse)
def is_service_healthy():
    """
    This endpoint is intended to determine whether the application container is at a healthy state
    or failed to startup up hence needs to be restarted.
    """
    return {"is_healthy": True}


@health_router.get("/readyz", status_code=200, response_model=ReadyResponse)
def is_service_ready(response: Response):
    """
    This endpoint is intended to determine whether the application is ready to accept requests from
    downstream services.
    """

    is_ready = True
    is_dashboard_available = os.path.isdir(UI_BUILD_PATH)

    if not is_dashboard_available:
        is_ready = False

    if not is_ready:
        response.status_code = status.HTTP_503_SERVICE_UNAVAILABLE

    return {"is_ready": is_ready, "is_dashboard_available": is_dashboard_available}



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.


import traceback
from datetime import datetime, timezone
from multiprocessing import Process
from multiprocessing import Queue as MPQ
from typing import Dict, List

import cloudpickle as pickle
import requests
from app.core.get_svc_uri import DispatcherURI, RunnerURI

from covalent._results_manager.result import Result

from .runner_logger import logger


def generate_task_result(
    task_id,
    start_time=None,
    end_time=None,
    status=None,
    output=None,
    error=None,
    stdout=None,
    stderr=None,
    info=None,
):

    return {
        "task_id": task_id,
        "start_time": start_time,
        "end_time": end_time,
        "status": status,
        "output": output,
        "error": error,
        "stdout": stdout,
        "stderr": stderr,
        "info": info,
    }


def send_task_update_to_dispatcher(dispatch_id, task_result):

    url = DispatcherURI().get_route(f"workflow/{dispatch_id}")
    response = requests.put(url=url, files={"task_execution_results": pickle.dumps(task_result)})
    response.raise_for_status()

    logger.warning(f"{dispatch_id}")
    logger.warning(f"{task_result}")


def free_resources_call_to_runner(dispatch_id, task_id):

    url = RunnerURI().get_route(f"workflow/{dispatch_id}/task/{task_id}/free")
    response = requests.post(url=url)
    response.raise_for_status()


def done_callback_to_runner(dispatch_id, task_id):

    url = RunnerURI().get_route(f"workflow/{dispatch_id}/task/{task_id}/done")
    response = requests.post(url=url)
    response.raise_for_status()


def start_task(task_id, func, args, kwargs, executor, results_dir, info_queue, dispatch_id):

    # These objects were pickled (with cloudpickle) in run_tasks_with_resources so as to be
    # compatible (pickleable) with multiprocessing processes.
    args = pickle.loads(args)
    kwargs = pickle.loads(kwargs)
    executor = pickle.loads(executor)

    task_result = generate_task_result(
        task_id=task_id,
        start_time=datetime.now(timezone.utc),
        status=Result.RUNNING,
    )

    # Set task as running and send update to dispatcher
    send_task_update_to_dispatcher(dispatch_id, task_result)

    logger.warning(f"info queue when inside the process {info_queue}")

    task_output, stdout, stderr, exception = executor.execute(
        function=func,
        args=args,
        kwargs=kwargs,
        info_queue=info_queue,
        task_id=task_id,
        dispatch_id=dispatch_id,
        results_dir=results_dir,
    )

    task_result = generate_task_result(
        task_id=task_id,
        end_time=datetime.now(timezone.utc),
        status=Result.FAILED if exception else Result.COMPLETED,
        output=task_output,
        error="".join(traceback.TracebackException.from_exception(exception).format())
        if exception
        else None,
        stdout=stdout,
        stderr=stderr,
        info=info_queue.get(),
    )

    # No more info needs to be stored about execution
    info_queue.close()

    # Free resources callback to runner
    free_resources_call_to_runner(dispatch_id, task_id)

    # Set task as complete and send update to dispatcher
    send_task_update_to_dispatcher(dispatch_id, task_result)

    # Callback to the runner to close this process
    done_callback_to_runner(dispatch_id, task_id)


def run_tasks_with_resources(
    dispatch_id: str,
    tasks_left_to_run: List[Dict],
    resources: MPQ,
    ultimate_dict: dict,
):
    # Example task:
    # {
    #    "task_id": 3,
    #    "func": Callable,
    #    "args": [1, 2, 3],
    #    "kwargs": {"a": 1, "b": 2},
    #    "executor": Executor,
    #    "results_dir": "/path/to/results/"
    # }

    if not ultimate_dict.get(dispatch_id):
        ultimate_dict[dispatch_id] = {}

    # Get number of available resources
    available_resources = resources.get()
    resources.put(available_resources)

    while available_resources > 0 and tasks_left_to_run:

        # Reduce the number of available resources
        resources.put(resources.get() - 1)

        # Popping first element from tasks_left_to_run
        task = tasks_left_to_run.pop(0)
        info_queue = MPQ()

        logger.warning(f"info queue when starting the process {info_queue}")

        # Organizing the args to be sent
        starting_args = (
            task["task_id"],
            task["func"],
            pickle.dumps(task["args"]),
            pickle.dumps(task["kwargs"]),
            pickle.dumps(task["executor"]),
            task["results_dir"],
            info_queue,
            dispatch_id,
        )

        process = Process(target=start_task, args=starting_args, daemon=True)
        process.start()

        ultimate_dict[dispatch_id][task["task_id"]] = {
            "process": process,
            "executor": pickle.dumps(task["executor"]),
            "info_queue": info_queue,
        }

        available_resources = resources.get()
        resources.put(available_resources)

    return tasks_left_to_run


def get_task_status(executor, info_queue):

    info = info_queue.get()
    status = executor.get_status(info)
    info_queue.put(info)

    return status


def cancel_running_task(executor, info_queue):

    # Using MPQ to get any information that execute method wanted to
    # share with cancel method
    executor.cancel(info_queue.get())

    # Close the info_queue for any more data transfer
    info_queue.close()
    info_queue.join_thread()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import logging
import shutil
import uuid
from abc import ABC
from pathlib import Path
from typing import BinaryIO, Generator, List, Union

from minio import Minio
from minio.error import S3Error

from .storagebackend import StorageBackend

logger = logging.getLogger(__name__)


class MinioStorageBackend(ABC):
    """Thin wrapper for Minio client

    Attributes:
        client: Minio client instance
        bucket_name: current working bucket name (default: "default")
    """

    def __init__(self, client, bucket_name="default"):
        self.bucket_name = bucket_name
        self.client = client

    def get(self, bucket_name: str, object_name: str) -> Union[Generator[bytes, None, None], None]:
        """Get object from storage.

        Args:
            bucket_name: name of the bucket
            object_name: name of the object

        Returns:
            A generator yielding a byte stream or None if an
            error occurred in retrieving the object

        """

        try:
            resp = self.client.get_object(bucket_name, object_name)
        except Exception as e:
            # TODO: better logging
            logger.debug("Exception in MinioStorageBackend:")
            logger.debug(e)
            return None

        if resp.status == 200:
            return resp.stream()
        else:
            return None

    def put(
        self,
        data: BinaryIO,
        bucket_name: str,
        object_name: str,
        length: int,
        metadata: dict = None,
        overwrite: bool = False,
    ) -> (str, str):

        """Upload object to storage.

        Args:
            data: a binary file-like object
            bucket_name: name of the bucket
            object_name: name of the destination object
            length: number of bytes to upload
            metadata: an optional dict of metadata to upload

        Returns:
            (bucket_name, object_name) if write succeeds and ("", "") otherwise

        """

        if not overwrite:
            try:
                res = self.client.stat_object(bucket_name, object_name)
                object_name = f"{uuid.uuid4()}.pkl"
            except S3Error as e:
                pass

        try:
            res = self.client.put_object(bucket_name, object_name, data, length, metadata=metadata)
        except Exception as e:
            # TODO: better logging
            logger.debug("Exception in MinioStorageBackend:")
            logger.debug(e)
            res = None

        if res:
            return (bucket_name, object_name)
        else:
            return ("", "")

    def delete(self, bucket_name: str, object_names: List[str]):
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the GNU Affero General Public License 3.0 (the "License").
# A copy of the License may be obtained with this software package or at
#
#      https://www.gnu.org/licenses/agpl-3.0.en.html
#
# Use of this file is prohibited except in compliance with the License. Any
# modifications or derivative works of this file must retain this copyright
# notice, and modified files must contain a notice indicating that they have
# been altered from the originals.
#
# Covalent is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the License for more details.
#
# Relief from the License may be granted by purchasing a commercial license.

import os

from furl import furl

from .config import _config_manager as cm


class ServiceURI:
    def __init__(
        self,
        scheme: str = "http",
        host: str = "localhost",
        port: int = None,
        prefix: str = "api/v0",
    ) -> None:
        self.scheme = scheme
        self.host = host
        self.port = port
        self.prefix = prefix

    def get_base_url(self):
        base_url = furl().set(scheme=self.scheme, host=self.host, port=self.port)
        if self.prefix:
            base_url.set(path=self.prefix)
        return base_url

    def get_route(self, path: str):
        base_url = self.get_base_url().copy()
        base_url.path /= path
        return base_url.url


class UIBackendURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(port=cm.get("UI_SVC_PORT"), host=cm.get("UI_SVC_HOST"))


class DispatcherURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(
            port=cm.get("DISPATCHER_SVC_PORT"),
            host=cm.get("DISPATCHER_SVC_HOST"),
        )


class QueuerURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(port=cm.get("QUEUER_SVC_PORT"), host=cm.get("QUEUER_SVC_HOST"))


class ResultsURI(ServiceURI):
    def __init__(self) -> None:
        super().__init__(
            port=cm.get("RESULTS_SVC_PORT"),
            host=cm.get("RESULTS_SVC_HOST"),
        )



import json
import os
import sqlite3
from typing import List, Tuple

from covalent._results_manager.result import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.utils import encode_result, get_named_params

# DB Schema:
# TABLE dispatches
# * dispatch_id text primary key
# * result_dict text ; json-serialied dictionary representation of Result

# TODO: Move these to a common utils module

app_log = logger.app_log


class DispatchDB:
    """
    Wrapper for the database of workflows.
    """

    def __init__(self, dbpath: str = None) -> None:
        if dbpath:
            self._dbpath = dbpath
        else:
            self._dbpath = get_config("user_interface.dispatch_db")

        # dispatch_id is the primary key

        # Initialize the db if necessary; sqlite3 raises
        # sqlite3.OperationalError if table already exists.
        self.conn = sqlite3.connect(self._dbpath)
        try:
            self.conn.execute(
                "CREATE TABLE dispatches \
                (dispatch_id text primary key, \
                result_dict text)"
            )
            self.conn.commit()
        except sqlite3.OperationalError:
            pass
        app_log.debug("Instantiated DispatchDB")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self.conn.rollback()
        else:
            self.conn.commit()
        self.conn.close()

        return False

    def get(self, dispatch_ids: [] = []) -> List[Tuple[str, str]]:
        """
        Retrieve workflows with the given dispatch ids.

        Args:
            dispatch_ids: A list of dispatch ids for the sought-after workflows.

        Returns:
            A list of pairs (dispatch_id, [jsonified result dictionary]).
        """
        if len(dispatch_ids) > 0:
            app_log.debug("DispatchDB.get() is retrieving results for dispatch ids.")
            app_log.debug(dispatch_ids)
            placeholders = "({})".format(", ".join(["?" for i in dispatch_ids]))
            sql = (
                "SELECT * FROM dispatches WHERE \
            dispatch_id in "
                + placeholders
            )

            res = self.conn.execute(sql, dispatch_ids).fetchall()
            app_log.debug("DispatchDB.get() is returning the following response object.")
            app_log.debug(res)

        else:
            sql = "SELECT * FROM dispatches"

            res = self.conn.execute(sql).fetchall()

        return res

    def upsert(self, dispatch_id: str, result_obj: Result) -> None:
        """
        Insert or update the record with the given dispatch_id.

        Args:
            dispatch_id: The workflow's dispatch_id.
            result_obj: The Result object for the workflow.

        The Result is turned into a dictionary and stored as json.
        """

        jsonified_result = encode_result(result_obj)

        try:
            sql = "INSERT INTO dispatches (dispatch_id, result_dict) VALUES (?, ?)"
            self.conn.execute(sql, (dispatch_id, jsonified_result))
            self.conn.commit()

        except sqlite3.IntegrityError:
            sql = "UPDATE dispatches SET result_dict = ? WHERE dispatch_id = ?"
            self.conn.execute(sql, (jsonified_result, dispatch_id))
            self.conn.commit()

        # sql = "INSERT INTO dispatches (dispatch_id, result_dict) VALUES (?, ?) \
        # ON CONFLICT (dispatch_id) DO UPDATE SET result_dict = excluded.result_dict"

        # self.conn.execute(sql, (dispatch_id, jsonified_result))

    def delete(self, dispatch_ids: []) -> None:
        """
        Delete records with the given dispatch ids.

        Args:
            dispatch_ids: A list of dispatch ids
        """
        placeholders = "({})".format(", ".join(["?" for i in dispatch_ids]))
        sql = "DELETE FROM dispatches WHERE dispatch_id in " + placeholders

        self.conn.execute(sql, dispatch_ids)
        self.conn.commit()
